import logging
try:
    from gevent import queue
except ImportError:
    from eventlet import queue

logger = logging.getLogger('django.geventpool')


class DatabaseConnectionPool:
    def __init__(self, maxsize=100):
        self.maxsize = maxsize
        self.pool = queue.Queue(maxsize=maxsize)
        self.size = 0

    def get(self, conn_params):
        if self.size >= self.maxsize or self.pool.qsize():
            conn = self.pool.get()
            if not self.is_usable(conn):
                try:
                    conn.close()
                except Exception:
                    pass
                conn = self.create_connection(conn_params)
            return conn
        else:
            self.size += 1
            try:
                conn = self.create_connection(conn_params)
            except Exception:
                self.size -= 1
                raise
            return conn

    def put(self, item):
        if item is None:
            self.size -= 1
            return
        try:
            self.pool.put(item, timeout=2)
        except queue.Full:
            item.close()

    def closeall(self):
        while not self.pool.empty():
            conn = self.pool.get_nowait()
            try:
                conn.close()
            except Exception:
                pass
        self.size = 0

    def create_connection(self, *args, **kwargs):
        raise NotImplementedError("create_connection")

    def is_usable(self, obj):
        raise NotImplementedError("is_usable")



import logging
try:
    from gevent.lock import Semaphore
except ImportError:
    from eventlet.semaphore import Semaphore

from django.db.backends.mysql.base import DatabaseWrapper as OriginalDatabaseWrapper
from .creation import DatabaseCreation
from .connection_pool import MysqlConnectionPool


logger = logging.getLogger('django.geventpool')

connection_pools = {}
connection_pools_lock = Semaphore(value=1)

DEFAULT_MAX_CONNS = 4


class DatabaseWrapper(OriginalDatabaseWrapper):
    creation_class = DatabaseCreation

    def __init__(self, settings_dict, *args, **kwargs):
        def pop_max_conn(settings_dict):
            if "OPTIONS" in settings_dict:
                return settings_dict["OPTIONS"].pop("MAX_CONNS", DEFAULT_MAX_CONNS)
            else:
                return DEFAULT_MAX_CONNS
        self._pool = None
        settings_dict['CONN_MAX_AGE'] = 0
        self._max_cons = pop_max_conn(settings_dict)
        super(DatabaseWrapper, self).__init__(settings_dict, *args, **kwargs)
        self.allow_thread_sharing = True

    @property
    def pool(self):
        if self._pool is not None:
            return self._pool
        connection_pools_lock.acquire()
        if self.alias not in connection_pools:
            self._pool = MysqlConnectionPool(self._max_cons)
            connection_pools[self.alias] = self._pool
        else:
            self._pool = connection_pools[self.alias]
        connection_pools_lock.release()
        return self._pool

    def get_new_connection(self, conn_params):
        if self.connection is None:
            self.connection = self.pool.get(conn_params)
            self.closed_in_transaction = False
        return self.connection

    def _close(self):
        if self.connection is None:
            self.pool.closeall()
        else:
            with self.wrap_database_errors:
                if not self.in_atomic_block:
                    self.pool.put(self.connection)
                else:
                    self.pool.put(None)
                    self.connection.close()

    def closeall(self):
        for pool in connection_pools.values():
            pool.closeall()



from functools import wraps

from django.core.signals import request_finished


def close_connection(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        finally:
            request_finished.send(sender='greenlet')
    return wrapper



from django.core.exceptions import ImproperlyConfigured
try:
    import MySQLdb as Database
except ImportError as err:
    raise ImproperlyConfigured(
        'Error loading MySQLdb module.\n'
        'Did you install mysqlclient? or install PyMySQL as MySQLdb'
    ) from err
from ..connection_pool import DatabaseConnectionPool


class MysqlConnectionPool(DatabaseConnectionPool):
    def __init__(self, maxsize):
        super(MysqlConnectionPool, self).__init__(maxsize)

    def create_connection(self, conn_params):
        return Database.connect(**conn_params)

    def is_usable(self, conn):
        try:
            conn.ping()
        except Database.Error:
            return False
        else:
            return True



from setuptools import setup, find_packages

setup(
    name='django-mysql-geventpool',
    version='0.2.1',
    install_requires=['django>=2.0', 'gevent'],
    description='Add a MySQL connection pool for django using gevent',
    long_description=open("README.md").read(),
    packages=find_packages(),
    include_package_data=True,
    license='Apache 2.0',
    lassifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
    author='aihara',
    author_email='aihara@argmax.jp'
)
